#!/usr/bin/perl

# Copyright (C) 2003-2005 Damien Krotkine (dams@libconf.net)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; only version 2, no any other version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

=head1 NAME

Libconf::Conf2XML - Config file to XML module

=head1 DESCRIPTION

Libconf::Conf2XML is a class that implements the connection between
configuration structure and XML. It provides simple methods to get XML from a
config file, or to get a config file from an XML string.

Note that this package is not really intended to be used directly. It's there
mainly for the C binding. If you want to interact with configuration files and
XML, see L<Libconf::Glueconf>, it has built-in XML bindings.

=head1 Config2XML METHODS

=over

=item B<conf2xml()>

  use Libconf::Conf2XML;
  $xml_string = Libconf::Conf2XML::conf2xml('/etc/password');
  $xml_string = Libconf::Conf2XML::conf2xml('/etc/conf.d/pure-ftpd', 'Generic/Shell');
  $xml_string = Libconf::Conf2XML::conf2xml('/etc/make.conf', 'Generic/Shell', 'shell_style:true_bash,shell_command:/bin/bash');

This method returns the xml_string from a filename of a config file. The second
and third arguments are optional, they represents the template and the options
to pass to the template. These arguments are useful if the filename is not
referenced in the internal database.

=cut

package Libconf::Conf2XML;
use Libconf qw(:helpers);

sub conf2xml {
    my ($filename, $template, $template_options) = @_;
    my $obj = _get_template($filename, $template, $template_options);
    my $xml_string;

    eval {
        $xml_string = $obj->toXMLString();
    };  $@ and die_error($@);
    return $xml_string;
}

sub die_error {
    my ($errmsg) = @_;
    chomp $errmsg;
    die ("an error occured, check the following message (if any) : \`\`$errmsg''\n\n");
}

=item B<xml2conf()>

  use Libconf::Conf2XML;
  $xml_string = Libconf::Conf2XML::xml2conf($filename, $xml_string);
  $xml_string = Libconf::Conf2XML::xml2conf($filename, $xml_string, 'Generic/Shell');
  $xml_string = Libconf::Conf2XML::xml2conf($filename, $xml_string, 'Generic/Shell', 'shell_style:true_bash,shell_command:/bin/bash');

This method takes the xml_string in a XML::Dumper format, generates the
configuration file and saves it to $filename. See conf2xml for explanation on
the other arguments.

=cut

sub xml2conf {
    my ($xml_string, $filename, $template, $template_options) = @_;
    if ($template eq '') {
        can_XML_Twig() or die "your version of libconf was not installed with xml support\n.XML::Twig is needed.\n";
        require XML::Twig;
        import XML::Twig;
        my $t= XML::Twig->new();
        $t->parse($xml_string);
        my $root= $t->root;
        defined $root && $root->tag() eq 'configfile' and $template = $root->att('template');
    }
    $template eq '' and die "no template specified, no template found in the xml. Please specify one.\n";
    my $obj = _get_template($filename, $template, $template_options, 1);
    eval {
        $obj->fromXMLString($xml_string);
        $obj->write_conf($filename);
    };  $@ and die_error($@);
    return 0;
}

sub _get_template_options {
    my ($template_options_arg) = @_;
    join (', ', map { my ($k, $v) = split(/:/, $_); qq($k => "$v"); } split(/,/, $template_options_arg));
}

sub _get_template {
    my ($filename, $template, $template_options, $dont_parse_file) = @_;
    my $template_orig = $template;
    $template =~ s|/|::|g;
    my $template_class = "Libconf::Glueconf::$template";
    my $obj;
    eval {
        $template_options2 = _get_template_options($template_options);
    }; $@ and die_error($@);
    eval qq| require $template_class; import $template_class; \$obj=new $template_class({ filename => "| . ($dont_parse_file ? "''" : $filename) . q|", $template_options2 });|;
    $@ and die_error($@);
    defined $obj or die("an error occured while creating the template, check the template name and the options (if any).\n");
    $obj->set_option(template_options => $template_options);
    $obj->set_option(show_comments => 0); # don't use this options, we'll manage comments ourselves
    $obj->set_option(glueconf_template_name => $template_orig); # set the template name
    $dont_parse_file and $obj->set_option(filename => $filename);
    $obj;
}

1;

=back

=cut
