#!/usr/bin/perl

# Author : Damien Krotkine (dams@gentoo.org)
#
# Copyright (C) 2002-2005 Damien Krotkine (dams@gentoo.org)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

package Libconf::Glueconf;
use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK);
use Libconf qw(:helpers);

@ISA = qw(Exporter);
@EXPORT = qw(init_args);

=head1 NAME

Libconf::Glueconf - Libconf high level common module

=head1 DESCRIPTION

Libconf::Glueconf is a class that represents a config file at a high level. It
provides an easy-to-use and powerful structure to edit the config file.

=head1 SYNOPSYS

=head2 PERL

This is the high level layer of libconf. You should use it to read/modify/write the
informations to the config files. Here are some examples, but you should look
in the specific modules below for more examples.

  use Libconf::Glueconf::Samba::Smb_conf;
  $samba = new Libconf::Glueconf::Samba::Smb_conf({ filename => '/etc/samba/smb.conf' });
  $samba->{homes}->{writable} = 'TEST';
  $samba->{global}->{workgroup} eq 'WORKGROUP' or die "workgroup is not correct\n";
  $samba->write_conf(); # write back to /etc/samba/smb.conf
  $samba->write_conf('/tmp/smb.conf'); # write to an alternate file 


  use Libconf::Glueconf::X::XF86Config;
  use Data::Dumper;
  $Data::Dumper::Deepcopy = 1;
  $Data::Dumper::Quotekeys = 0;
  $xorg = new Libconf::Glueconf::X::XF86Config({ filename => '/etc/X11/xorg.conf'} );
  print Dumper($xorg);

  use Libconf::Glueconf::System::Inittab;
  $inittab = new Libconf::Glueconf::System::Inittab({ filename => '/etc/inittab' });
  $inittab->{id}->{runlevels} = '3';
  $inittab->write_conf();

=head2 C, PYTHON, RUBY

Use the following module list, take the last 2 words, and replace :: with /
Example : Libconf::Glueconf::Generic::KeyValue will become 'Generic/KeyValue'.
Use it as arguments to Conf2Xml.conf2xml (ruby), conf2xml (python), conf2xml
(c). More details available here (XXX TODO need to be written)

=head1 MODULE LIST

Here is the list of available templates, see their documentation for specific
methods

=head2 Generic templates

You should use these if there is no specific templates fitting your needs. Look
at their documentation to get the options you caan use to change their
behaviour

=over

=item B<Generic - KeyValue>

L<Libconf::Glueconf::Generic::KeyValue> : [ key - value ] atoms.

=item B<Generic - KeyValues>

L<Libconf::Glueconf::Generic::KeyValues> : [ key - value1,value2 ] atoms.

=item B<Generic - ValuesSection>

L<Libconf::Glueconf::Generic::ValueSection> : [ Section ] with [ Value ] atoms.

=item B<Generic - Shell>

L<Libconf::Glueconf::Generic::Shell> : shell style configuration files.

=back

=head2 Specific templates

These templates are made to parse precise configuration files.

=over 

=item B<NUT - Ups_conf>

L<Libconf::Glueconf::NUT::Ups_conf> : template for UPS hardwares configuration files

=item B<Networking - Hosts>

L<Libconf::Glueconf::Networking::Hosts> : template for /etc/hosts file

=item B<Networking - Resolv>

L<Libconf::Glueconf::Networking::Resolv> : template for /etc/resolv.conf file

=item B<Networking - Sshd_config>

L<Libconf::Glueconf::Networking::Sshd_config> : template for ssh deamon config file

=item B<Postfix - Main_cf>

L<Libconf::Glueconf::Postfix::Main_cf> : template for postfix main.cf like files

=item B<Samba - Smb_conf>

L<Libconf::Glueconf::Samba::Smb_conf> : template for /etc/samba/smb.conf like files

=item B<X - XF86Config>

L<Libconf::Glueconf::X::XF86Config> : template for xorg.conf like files

=item B<X - Xdm>

L<Libconf::Glueconf::X::Xdm> : template for xdm configuration file

=item B<X - Gdm>

L<Libconf::Glueconf::X::Gdm> : template for gdm and kdm configuration files

=back

=head1 GLUECONF GENERAL METHODS (PERL ONLY)

=over

=item B<read_conf()>

  use Data::Dumper;
  my $structure = new Libconf::Glueconf::Some::Module;
  $structure->read_conf();
  print Dumper($structure) . "\n";

B<description>

This method reads the config file, and maps it to the structure. It doesn't use
any argument for its own use, but you can give it arguments, they will be
passed to the underlying Libconf::Templates::read_conf().

=cut

sub read_conf { shift->{libconf}->read_conf(@_) }

=item B<write_conf()>

  my $structure = new Libconf::Glueconf::Some::Module;
  $structure->read_conf();
  ... edit $structure ...
  $structure->write_conf();

B<description>

This method writes the structure back to the config file. It doesn't use any
argument for its own use, but you can give it arguments, they will be passed to
the underlying Libconf::Templates::write_conf().

=cut

sub write_conf { shift->{libconf}->write_conf(@_) }

=item B<set_option($key, $value)>

    $template->set_option(filename => '/etc/test.cfg');

B<description>

sets an option after instanciation of the object.

=cut

sub set_option {
    my ($obj, $key, $value) = @_;
    $obj->{libconf}{$key} = $value;
}

=item B<get_option($key)>

    my $value = $template->get_option('filename');

B<description>

get the value of an option after instanciation of the object.

=cut

sub get_option {
    my ($obj, $key) = @_;
    $obj->{libconf}{$key};
}

=item B<toXMLString($format)>

    $template->toXMLString();

B<description>

toXMLString exports the template informations into an XML string.

The optional $format parameter sets the indenting of the output.

=cut

sub to_xml_loop {
    my ($branch, $obj, $path, $root_obj) = @_;
    if ($path ne '') {
        debug("****** " . "comments_struct = root_obj->get_comments($path)");
        my $comments_struct = $root_obj->get_comments($path);
        my $comments;
        foreach (qw(normal left_inline right_inline)) {
            if ($comments_struct->{$_} ne '') {
                $comments ||= new XML::Twig::Elt('comments');
                new XML::Twig::Elt($_, $comments_struct->{$_})->paste('last_child', $comments)
            }
        }
        defined $comments and $comments->paste('last_child', $branch);
    }

    if (UNIVERSAL::isa($obj, 'ARRAY')) {
        my $array = new XML::Twig::Elt('array');
        my $i = 0;
        foreach (@$obj) {
            my $item = new XML::Twig::Elt(item => { key => $i });
            to_xml_loop($item, $_, $path . "[$i]", $root_obj);
            $item->paste('last_child', $array);
        } continue { $i++; }
        $array->paste('last_child', $branch);
    } elsif (UNIVERSAL::isa($obj, 'HASH')) {
#         if (exists $obj->{_value} || exists $obj->{_comments}) {
#             my $value = new XML::Twig::Elt(value => $obj->{_value});
#             $value->paste('last_child', $branch);
#             if (defined $obj->{_comments}) {
#                 if (join('', map { $obj->{_comments}{$_} } qw(right_inline left_inline normal)) ne '') {
#                     my $comments = new XML::Twig::Elt('comments');
#                     foreach (qw(right_inline left_inline normal)) {
#                         $obj->{_comments}{$_} ne '' and new XML::Twig::Elt($_, $obj->{_comments}{$_})->paste('last_child', $comments);
#                     }
#                     $comments->paste('last_child', $branch);
#                 }
#             }
#         } else {
        my $hash = new XML::Twig::Elt('hash');
        foreach (keys(%$obj)) {
            my $item = new XML::Twig::Elt(item => { key => $_ });
            to_xml_loop($item, $obj->{$_}, $path . '{"' . quotemeta($_) . '"}', $root_obj);
            $item->paste('last_child', $hash);
        }
        $hash->paste('last_child', $branch);
#        }
    } elsif (UNIVERSAL::isa($obj, 'SCALAR')) {
        my $scalarref = new XML::Twig::Elt('scalarref');
        to_xml_loop($scalarref, $$obj, $path, $root_obj);
        $scalarref->paste('last_child', $branch);
    } else {
        new XML::Twig::Elt(value => $obj)->paste('last_child', $branch);
    }
}

sub toXMLString {
    my ($obj, $format) = @_;
    can_XML_Twig() or die "your version of libconf was not installed with xml support\n.XML::Twig is needed.\n";

    require XML::Twig;
    import XML::Twig;

    my $t= XML::Twig->new();
    $t->parse('<configfile></configfile>');
    my $root= $t->root;
    to_xml_loop($root, $obj, '', $obj);

    $root->set_att(filename => $obj->get_option('filename'));
    $root->set_att(template => $obj->get_option('glueconf_template_name'));
    $root->set_att(options => $obj->get_option('template_options'));

    get_xml_header() . $t->sprint(pretty => 1);
}

sub get_xml_header {

# array -> array ref
# hash -> array ref
# value -> scalar ref

#  <!ELEMENT item ((value,comments?)|((#PCDATA|scalarref|array|hash)*))>

q(<?xml version="1.0"?>
<!DOCTYPE configfile [
  <!ELEMENT scalarref ((comments)?,(value)?)>
  <!ELEMENT array ((comments)?,item*)>
  <!ELEMENT hash ((comments)?,item*)>

  <!ELEMENT value (#PCDATA)>
  <!ELEMENT comments ((left_inline)?,(right_inline)?,(normal)?)>

  <!ELEMENT left_inline (#PCDATA)>
  <!ELEMENT right_inline (#PCDATA)>
  <!ELEMENT normal (#PCDATA)>

  <!ELEMENT item ((comments)?,(value|scalarref|array|hash)?)>
  <!ATTLIST item 
        key CDATA #REQUIRED>
  <!ATTLIST configfile
        filename CDATA #IMPLIED
        template CDATA #IMPLIED
        options  CDATA #IMPLIED>
  <!ELEMENT configfile ((array|hash)*)>
]>);
}


=item B<toXMLFile($filename, $format)>

    $template->toXMLFile('xml_filename');

B<description>

toXMLFile exports the template informations into an XML file.

The optional $format parameter sets the indenting of the output.

=cut

sub toXMLFile {
    my ($obj, $filename, $format) = @_;
    can_DumpXML() or die "your version of Libconf was not installed with xml dumping support.\nXML::Dumper is needed.\n";
    output($filename, toXMLString($obj, $format));
}

=item B<fromXMLString($xml_string)>

    $template->fromXMLString($xml_string);

B<description>

fromXMLString imports the data from an XML string in the format previously generated by toXMLString

=cut

sub debug { use Libconf; Libconf::debug(@_) }

my %type2path = (hash => '{"%s"}', array => '[%s]');
sub from_xml_command {
    my ($obj, $command)= @_;
    debug($command);
    eval ($command);
    $@ and die "unexpected error while evaluating '$command' (error was: $@)";    
}

sub from_xml_loop {
    my ($branch, $obj, $path, $type, $command_list) = @_;
    my @children = $branch->children();
    if (@children == 0) {
#        debug("$path=undef;\n");
#        eval ("$path=undef;");
#        $@ and die "unexpected error while evaluating '$path=undef;' (error was: $@)";
    } else {
        foreach my $child (@children) {
            my $tag = $child->tag();
            if ($tag eq 'comments') {
                debug("***********IN COMMENT");
                my $comments_txt = '{';
                foreach my $granchild ($child->children()) {
                    foreach (qw(normal left_inline right_inline)) {
                        $granchild->tag() eq $_ and $comments_txt .= qq($_ => ") .  quotemeta($granchild->text) . '",';
                    }
                }
                $comments_txt .= '}';
                $path =~ /\$obj->(.*)$/;
                push @$command_list, q|$obj->set_comments('| . $1 . qq|', $comments_txt);|;
            } elsif ($tag eq 'value') {
                from_xml_loop($child, $obj, $path, $command_list);
            } elsif ($tag eq '#PCDATA') {
                debug("***********IN PCDATA");
                from_xml_command($obj, qq($path=") . quotemeta($branch->text) . q(";));
            } elsif ($tag eq 'hash') {
                debug("***********IN HASH");
                from_xml_command($obj, 'defined ' . $path . ' or %{' . $path . '}=();');
                from_xml_loop($child, $obj, $path, 'hash', $command_list);
            } elsif ($tag eq 'array') {
                debug("***********IN ARRAY");
                from_xml_command($obj, 'defined ' . $path . ' or @{' . $path . '}=();');
                from_xml_loop($child, $obj, $path, 'array', $command_list);
            } elsif ($tag eq 'scalarref') {
                debug("***********IN SCALARREF");
                from_xml_loop($child, $obj, "\${$path}", '', $command_list);
            } elsif ($tag eq 'item') {
                debug("***********IN ITEM");
                my $key = $child->att('key');
                from_xml_loop($child, $obj, $path . '->' . sprintf($type2path{$type}, quotemeta($key)), '', $command_list);
            }
        }
    }
}

sub fromXMLString {
    my ($obj, $xml_string) = @_;
    can_XML_Twig() or die "your version of libconf was not installed with xml support\n.XML::Twig is needed.\n";

    require XML::Twig;
    import XML::Twig;

    my $t= XML::Twig->new();
    $t->parse($xml_string);
    my $root= $t->root;
    my @command_list = ();
    from_xml_loop($root, $obj, '$obj', '', \@command_list);
    foreach my $command (@command_list) {
        from_xml_command($obj, $command);
    }
    
}

=item B<fromXMLFile($filename)>

    $template->fromXMLFile('file.xml');

B<description>

fromXMLFile imports the data from an XML file previously generated by toXMLFile

=cut

sub fromXMLFile {
    my ($obj, $filename) = @_;
    fromXMLString(scalar(cat_($filename)));
}

=item B<get_comments('path description')>

B<description> 

get the comments of an atom, described by the structure.

B<example>

$comments = $samba->get_comments('{global}{workgroup}')

=cut

sub get_comments {
    my ($obj, $path) = @_;
    $obj->{libconf}{_fetch_atom_pos} = 1;
    $obj->{libconf}{_atom_pos} = undef;
    eval('$obj->' . $path);
    $@ and die $@;
    $obj->{libconf}{_fetch_atom_pos} = 0;
    $obj->{libconf}->get_comments($obj->{libconf}{_atom_pos});
}

=item B<set_comments('path description', $comments)>

B<description> 

set the comments of an atom, described by the structure.

B<example>

$samba->set_comments('{global}{workgroup}', $comments)

=cut

sub set_comments {
    my ($obj, $path, $comments) = @_;
    $obj->{libconf}{_fetch_atom_pos} = 1;
    $obj->{libconf}{_atom_pos} = undef;
    eval('$obj->' . $path);
    $@ and die $@;
    $obj->{libconf}{_fetch_atom_pos} = 0;
    $obj->{libconf}->set_comments($obj->{libconf}{_atom_pos}, $comments);
}

=item B<set_commented('path description', $value)>

B<description> 

change the status of the atom to be commented or uncommented. A commented line will output commented with the first comment of $self->{comment_struct}

B<example>

$samba->set_commented('{global}{workgroup}', 1)

when outputed, the line will be something like :
# workgroup=foo;

=cut

sub set_commented {
    my ($obj, $path, $commented) = @_;
    $obj->{libconf}{_fetch_atom_pos} = 1;
    my $atom_pos;
    eval('$atom_pos = $obj->' . $path);
    $@ and die $@;
    $obj->{libconf}{_fetch_atom_pos} = 0;
    defined $atom_pos or return undef;
    my $atom = $obj->{libconf}->get_atom($atom_pos);
    defined $atom or return undef;
    $atom->{commented} = $commented;
}

=item B<commented('path description')>

B<description> 

returns the comment status of the atom.

B<example>

if ($samba->commented('{global}{workgroup}')) {
...
}

=cut

sub commented {
    my ($obj, $path) = @_;
    $obj->{libconf}{_fetch_atom_pos} = 1;
    my $atom_pos;
    eval('$atom_pos = $obj->' . $path);
    $@ and die $@;
    $obj->{libconf}{_fetch_atom_pos} = 0;
    defined $atom_pos or return undef;
    my $atom = $obj->{libconf}->get_atom($atom_pos);
    defined $atom or return undef;
    $atom->{commented};
}

#Warning, duplicated from Templates;
sub init_args {
    my ($args, $default_args) = @_;
    $args ||= {};
    add2hash_({ %$args }, $default_args);
}

# dummy function, should be cleaned up

sub bool2text { $_[0] ? "true" : "false" }
sub bool2yesno { $_[0] ? "yes" : "no" }
sub text2bool { my $t = lc($_[0]); $t eq "true" || $t eq "yes" ? 1 : 0 }


# dummy functions for compatibility

sub readConf {
    warn "readConf is OBSOLETE, you should use read_conf now. readConf will be removed soon.";
    shift->read_conf(@_);
}

sub writeConf {
    warn "writeConf is OBSOLETE, you should use write_conf now. writeConf will be removed soon.";
    shift->write_conf(@_);
}

=back

=cut

1;
